package ui.controller;

import classes.*;
import classes.comparator.EnergyMinToMaxComparator;
import classes.comparator.MinEnergyComparator;
import classes.comparator.WeakestBattery;
import ui.model.CableWithState;
import ui.model.DecoratedCable;
import ui.model.DecoratedCable.CableState;
import ui.model.DecoratedSwitch.SwitchState;
import ui.model.DecoratedNetwork;
import ui.model.DecoratedState;
import ui.model.DecoratedSwitch;
import ui.model.MinimumModel;
import ui.model.MinimumNetwork;
import ui.model.Model;
import ui.model.Model.FairnessModel;
import ui.model.VisualRepresentationalState;
import ui.view.FlexiblePane;
import ui.view.GUI;
import ui.view.MyCanvas;
import ui.view.Outliner;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.ListIterator;

import javax.swing.JPanel;


/**
 * Controller for Simulation.
 *
 * @author Gruppe14
 */
public class SimulationManager {
	int global = 0;
	private Model model;
	private ArrayList<AbstractCpsObject> objectsToHandle;
	private ArrayList<SubNet> subNets;
	private ArrayList<CpsEdge> brokenEdges;
	private HashMap<Integer, DecoratedState> saves = new HashMap<Integer, DecoratedState>();
	private HashMap<Integer, VisualRepresentationalState> savesVisual = new HashMap<Integer, VisualRepresentationalState>();
	private MyCanvas canvas;
	private int timeStep;
	private HashMap<Integer, Float> tagTable = new HashMap<>();
	private FlexiblePane flexPane;
	private GUI gui;
	private HashMap<HolonElement, Float> flexDevicesTurnedOnThisTurn = new HashMap<>();

	/**
	 * Constructor.
	 *
	 * @param m
	 *            Model
	 */
	SimulationManager(Model m) {
		canvas = null;
		model = m;
		subNets = new ArrayList<>();
		brokenEdges = new ArrayList<>();
	}

	/**
	 * calculates the flow of the edges and the supply for objects and consider old timesteps for burned cables.
	 *
	 * @param timestep
	 *            current Iteration
	 */
	void calculateStateForTimeStep(int timestep) {
		
		
		HashMap<CpsEdge, CableState> map = new HashMap<CpsEdge, CableState>();
		if(timestep > 0 && saves.containsKey(timestep-1)) //if the state before exist
		{
			//make cable hastmap
			DecoratedState theStateBefore = saves.get(timestep-1);
			for(DecoratedCable edge : theStateBefore.getLeftOverEdges())
			{
				map.put(edge.getModel(), edge.getState());
			}
		}
		timeStep = timestep;
		ArrayList<MinimumNetwork> list =  new ArrayList<MinimumNetwork>();
		MinimumModel minimumModel = new MinimumModel(model.getObjectsOnCanvas(), model.getEdgesOnCanvas());
		//set all working:
		for(CableWithState cable : minimumModel.getEdgeList()) {
			if(map.containsKey(cable.getModel())) cable.setState(map.get(cable.getModel()));
		}
		ArrayList<CableWithState> leftOver = new ArrayList<CableWithState>();
		boolean doAnotherLoop = true;
		while(doAnotherLoop) {
			doAnotherLoop = false;
			list = calculateNetworks(minimumModel, timestep, leftOver);
			for(MinimumNetwork net : list) {
				float energyOnCables = net.getHolonObjectList().stream().filter(object -> object.getEnergyAtTimeStep(timestep) > 0.0f).map(object -> object.getEnergyAtTimeStep(timestep)).reduce(0.0f, ((a,b) -> a + b));
				//find the cable with the energy supplied from his two connected objects are the biggest, from all cables that the network give more energy than the cablecapacity. 
				CableWithState cable = net.getEdgeList().stream().filter(aCable -> energyOnCables > aCable.getModel().getCapacity()).max((lhs,rhs) -> Float.compare(lhs.getEnergyFromConnetedAtTimestep(timestep), rhs.getEnergyFromConnetedAtTimestep(timestep))).orElse(null);
				if(cable != null) {
					cable.setState(CableState.Burned);
					doAnotherLoop = true;
				}
			}
		}
		ArrayList<DecoratedNetwork> decorNetworks = new ArrayList<DecoratedNetwork>();
		FairnessModel actualFairnessModel = model.getFairnessModel();
		for (MinimumNetwork net : list) {
			decorNetworks.add(new DecoratedNetwork(net, timestep, actualFairnessModel));
		}
		ArrayList<DecoratedCable> leftOverDecoratedCables = new ArrayList<DecoratedCable>();
		
		for(CableWithState cable: leftOver) {
			leftOverDecoratedCables.add(new DecoratedCable(cable.getModel(), cable.getState(), 0.0f));
		}
		ArrayList<DecoratedSwitch> listOfDecoratedSwitches = decorateSwitches(minimumModel, timestep);
		DecoratedState stateFromThisTimestep = new DecoratedState(decorNetworks, leftOverDecoratedCables, listOfDecoratedSwitches, timestep);
		saves.put(timestep, stateFromThisTimestep);
		savesVisual.put(timestep, new VisualRepresentationalState(stateFromThisTimestep, minimumModel));
		canvas.repaint();
		gui.updateOutliners(getActualDecorState());//saves.getOrDefault(timestep, null);
		
	}
	/** 
	 * Decorate a switch
	 * @param minModel
	 * @param iteration
	 * @return
	 */
	public static ArrayList<DecoratedSwitch> decorateSwitches(MinimumModel minModel, int iteration) {
		ArrayList<DecoratedSwitch> aListOfDecoratedSwitches = new ArrayList<DecoratedSwitch>(); 
		for(HolonSwitch hSwitch: minModel.getSwitchList()) {
			aListOfDecoratedSwitches.add(new DecoratedSwitch(hSwitch, hSwitch.getState(iteration) ? SwitchState.Closed : SwitchState.Open));
		}
		return aListOfDecoratedSwitches;
	}
	/**
	 * SubFunction to calculate the Networks from the model.
	 * @param minModel
	 * @param Iteration
	 * @param leftOver
	 * @return
	 */
	ArrayList<MinimumNetwork> calculateNetworks(MinimumModel minModel, int Iteration, ArrayList<CableWithState> leftOver){
		//Copy minModel ObjectList
		ArrayList<HolonObject> holonObjectList = new ArrayList<HolonObject>();
		for(HolonObject holonObject: minModel.getHolonObjectList()) {
			holonObjectList.add(holonObject);
		}
		//Copy minModelEdgeList
		ArrayList<CableWithState> edgeList = new ArrayList<CableWithState>();
		for(CableWithState cable: minModel.getEdgeList()) {
			edgeList.add(cable);
		}
		
		ArrayList<MinimumNetwork> listOfNetworks = new ArrayList<MinimumNetwork>();
		while(!holonObjectList.isEmpty()) {
			//lookAt the first holonObject and find his neighbors
			HolonObject lookAtObject = holonObjectList.get(0);
			//delete out of list
			holonObjectList.remove(0);
			//create a new Network
			MinimumNetwork actualNetwork = new MinimumNetwork(new ArrayList<HolonObject>(), new ArrayList<CableWithState>());
			actualNetwork.getHolonObjectList().add(lookAtObject);
			//create List of neighbors
			LinkedList<AbstractCpsObject> neighbors = new LinkedList<AbstractCpsObject>();
			populateListOfNeighbors(edgeList, lookAtObject, actualNetwork, neighbors);
			while(!neighbors.isEmpty()) {
				AbstractCpsObject lookAtNeighbor = neighbors.getFirst();
				if(lookAtNeighbor instanceof HolonObject) {
					actualNetwork.getHolonObjectList().add((HolonObject) lookAtNeighbor);
					holonObjectList.remove(lookAtNeighbor);
				}
				//When HolonSwitch Check if closed
				if(!(lookAtNeighbor instanceof HolonSwitch) || ((HolonSwitch)lookAtNeighbor).getState(Iteration)) {
					populateListOfNeighbors(edgeList, lookAtNeighbor, actualNetwork, neighbors);
				}
				
				neighbors.removeFirst();
			}
			listOfNetworks.add(actualNetwork);
		}	
		if(leftOver!= null) {
			leftOver.clear();
			for(CableWithState cable: edgeList) {
				leftOver.add(cable);
			}
		}
		return listOfNetworks;
	}
	/**
	 * Adds the neighbors.
	 * @param edgeList
	 * @param lookAtObject
	 * @param actualNetwork
	 * @param neighbors
	 */
	void populateListOfNeighbors(ArrayList<CableWithState> edgeList, AbstractCpsObject lookAtObject,
			MinimumNetwork actualNetwork, LinkedList<AbstractCpsObject> neighbors) {
		ListIterator<CableWithState> iter = edgeList.listIterator();
		while(iter.hasNext())
		{
			CableWithState lookAtEdge = iter.next();
			if(lookAtEdge.getState() == CableState.Working && lookAtEdge.getModel().isConnectedTo(lookAtObject)) {
				iter.remove();
				actualNetwork.getEdgeList().add(lookAtEdge);
				
				//Add neighbar
				AbstractCpsObject edgeNeighbor;
				if(lookAtEdge.getModel().getA().equals(lookAtObject)) {
					edgeNeighbor = lookAtEdge.getModel().getB();
					
				}else {
					edgeNeighbor = lookAtEdge.getModel().getA();
				}
				if(!neighbors.contains(edgeNeighbor)) {
					neighbors.add(edgeNeighbor);
				}
			}
		}
	}
	/**
	 * add all battries.getOut() from a list of battries and return them
	 * @param aL a List of HolonBattries likely from subnet.getBatteries()
	 * @param x TimeStep
	 * @return 
	 * 
	 */
	private float GetOutAllBatteries(ArrayList<HolonBattery> aL, int x)
	{
		float OutEnergy = 0;
		for(HolonBattery hB : aL)
		{
			//System.out.println("Iteration: "+ x +"OutBattery: "+ hB.getOutAtTimeStep(x-1));
			OutEnergy += hB.getOutAtTimeStep(x-1);
		}
		//System.out.println("Iteration: "+ x +"GetOutAllBatteries: "+ OutEnergy);
		return OutEnergy;
	}
	/**
	 * search for all flexible devices in the network and turn them on, until
	 * energy surplus = 0 or all devices have been examined.
	 *
	 * This code could be compressed (cases inside over- and underproduction are
	 * the same), but we decided that it is better readable this way
	 *
	 * @param subNet
	 *            the subnet
	 * @param energySurplus
	 *            the current surplus of energy
	 */
	private void turnOnFlexibleDevices(SubNet subNet, float energySurplus,
			int timestep) {
		for (HolonObject holonOb : subNet.getObjects()) {
			for (HolonElement holonEl : holonOb.getElements()) {

				// if this element is flexible and active (can be considered for
				// calculations)
				if (holonEl.isFlexible() && holonEl.isActive()) {
					float energyAvailableSingle = holonEl
							.getEnergyAtTimeStep(timestep);
					float energyAvailableMultiple = energyAvailableSingle
							* holonEl.getAmount();

					// ------------- flexible consumer / OVERPRODUCTION
					// -------------
					if (energyAvailableMultiple < 0 && energySurplus > 0) {

						// if there is more wasted energy than energy that this
						// device can give, give all energy available
						if (Math.abs(energyAvailableMultiple) <= Math
								.abs(energySurplus)) {
							energySurplus += energyAvailableMultiple;
							// set the new energy consumption to the maximum
							holonEl.setEnergyPerElement(energyAvailableSingle);
							flexDevicesTurnedOnThisTurn.put(holonEl,
									energyAvailableMultiple);
						}
						// else: we just need to turn on part of the flexible
						// energy available
						else {
							float energyNeeded = -energySurplus;
							energySurplus += energyNeeded; // should give 0, but
															// was kept this was
															// for consistency
							// the energy needed divided through the amount of
							// elements
							holonEl.setEnergyPerElement(energyNeeded
									/ holonEl.getAmount());
							flexDevicesTurnedOnThisTurn.put(holonEl,
									energyNeeded);
						}
					}

					// ------------- flexible producer / UNDERPRODUCTION
					// -------------
					else if (energyAvailableMultiple > 0 && energySurplus < 0) {

						// if there is more energy needed than this device can
						// give, give all the energy available
						if (Math.abs(energyAvailableMultiple) <= Math
								.abs(energySurplus)) {
							energySurplus += energyAvailableMultiple;
							// set the new energy consumption to the maximum
							holonEl.setEnergyPerElement(energyAvailableSingle);
							flexDevicesTurnedOnThisTurn.put(holonEl,
									energyAvailableMultiple);
						}
						// else: we just need to turn on part of the flexible
						// energy available
						else {
							float energyNeeded = -energySurplus;
							int i = 0;
							energySurplus += energyNeeded; // should give 0, but
															// was kept this was
															// for consistency
							// the energy needed divided through the amount of
							// elements
							holonEl.setEnergyPerElement(energyNeeded
									/ holonEl.getAmount());
							flexDevicesTurnedOnThisTurn.put(holonEl,
									energyNeeded);
						}

					}
				}

				if (energySurplus == 0) {
					break;
				}
			}

			if (energySurplus == 0) {
				break;
			}
		}
	}

	/**
	 * Set Flow Simulation.
	 *
	 * @param sN
	 *            Subnet
	 */
	private void setFlowSimulation(SubNet sN) {
		ArrayList<AbstractCpsObject> producers = new ArrayList<>();
		AbstractCpsObject tmp = null;
		tagTable = new HashMap<>();
		// traverse all objects in this subnet
		for (HolonObject hl : sN.getObjects()) {
			float energy = hl.getEnergyAtTimeStep(timeStep);
			// if their production is higher than their consumption
			if (energy > 0) {
				tagTable.put(hl.getId(), energy);
				hl.addTag(hl.getId());
				for (CpsEdge edge : hl.getConnections()) {
					if (edge.isWorking()) {
						// set other end of edge as tmp-object
						// and add this end to the other end's tag-list
						AbstractCpsObject a = edge.getA();
						AbstractCpsObject b = edge.getB();
						if (a.getId() == hl.getId()) {
							b.addTag(hl.getId());
							tmp = b;
						}
						if (b.getId() == hl.getId()) {
							a.addTag(hl.getId());
							tmp = a;
						}

						edge.setFlow(edge.getFlow() + energy);
						edge.calculateState();
						edge.addTag(hl.getId());
						if (edge.isWorking() && !producers.contains(tmp)) {
							if (tmp instanceof HolonSwitch) {
								if (((HolonSwitch) tmp).getState(timeStep)) {
									producers.add(tmp);
								}
							} else if (!(tmp instanceof CpsUpperNode)) {
								producers.add(tmp);
							}
						}
					}
				}
			}
		}
		setFlowSimRec(producers, 0);
	}

	/**
	 * Set Flow Simulation Rec.
	 *
	 * @param nodes
	 *            the nodes
	 * @param iter
	 *            the Iteration
	 */
	private void setFlowSimRec(ArrayList<AbstractCpsObject> nodes, int iter) {
		ArrayList<AbstractCpsObject> newNodes = new ArrayList<>();
		ArrayList<CpsEdge> changedEdges = new ArrayList<>();
		AbstractCpsObject tmp;
		if (nodes.size() != 0) {
			for (AbstractCpsObject cps : nodes) {
				// check whether the cps is in a legit state if it is a switch
				if (legitState(cps)) {
					for (CpsEdge edge : cps.getConnections()) {
						// is edge working
						// and does the edge's tag-list not (yet) contain all
						// tags of the cps
						if (edge.isWorking()
								&& (!(edge.containsTags(edge.getTags(),
										cps.getTag())))) {
							if (edge.getA().getId() == cps.getId()) {
								tmp = edge.getB();
							} else {
								tmp = edge.getA();
							}
							for (Integer tag : cps.getTag()) {
								if (!(edge.getTags().contains(tag))
										&& !(edge.getPseudoTags().contains(tag))) {
									edge.setFlow(edge.getFlow()
											+ tagTable.get(tag));
									edge.addTag(tag);
								}
							}
							// uppernodes do not spread energy
							if (!(tmp instanceof CpsUpperNode)) {
								for (Integer tag : tmp.getTag()) {
									if (!(edge.getTags().contains(tag))
											&& tagTable.get(tag) != null
											&& !(edge.getPseudoTags()
													.contains(tag))) {
										edge.setFlow(edge.getFlow()
												+ tagTable.get(tag));
										edge.addPseudoTag(tag);
										changedEdges.add(edge);
									}
								}
							}
							edge.calculateState();
							if (edge.isWorking()
									&& !(tmp instanceof CpsUpperNode)) {
								tmp.addAllPseudoTags(cps.getTag());
								if (!newNodes.contains(tmp)) {
									newNodes.add(tmp);
								}
							}
						}
					}
				}
			}
			setPseudoTags(newNodes, changedEdges);
			setFlowSimRec(newNodes, iter + 1);
		}
	}

	/**
	 * Set the Pseudo Tags.
	 *
	 * @param nodes
	 *            Array of AbstractCpsObjects
	 */
	private void setPseudoTags(ArrayList<AbstractCpsObject> nodes,
			ArrayList<CpsEdge> edges) {
		for (AbstractCpsObject node : nodes) {
			node.recalculateTags();
			node.setPseudoTags(new ArrayList<>());
		}
		for (CpsEdge edge : edges) {
			edge.recalculateTags();
			edge.setPseudoTag(new ArrayList<>());
		}
	}

	/**
	 * Reset the Connection.
	 *
	 * @param cps
	 *            CpsObject
	 * @param visitedObj
	 *            the visited Objects
	 * @param visitedEdges
	 *            the visited Edges
	 */
	private void resetConnections(AbstractCpsObject cps,
			ArrayList<Integer> visitedObj, ArrayList<CpsEdge> visitedEdges) {
		visitedObj.add(cps.getId());
		cps.resetTags();
		for (CpsEdge e : cps.getConnections()) {
			if (!(visitedEdges.contains(e))) {
				e.setFlow(0);
				e.calculateState();
				e.setTags(new ArrayList<>());
				visitedEdges.add(e);
				if (!(visitedObj.contains(e.getA().getId()))) {
					resetConnections(e.getA(), visitedObj, visitedEdges);
					e.getA().resetTags();
				}
				if (!(visitedObj.contains(e.getB().getId()))) {
					resetConnections(e.getB(), visitedObj, visitedEdges);
					e.getB().resetTags();
				}
			}
		}
	}

	/**
	 * calculates the energy of either all producers or consumers. Flexible
	 * devices are filtered out
	 *
	 * @param type
	 *            Type
	 * @param sN
	 *            Subnet
	 * @param x
	 *            Integer
	 * @return The Energy
	 */
	private float calculateEnergyWithoutFlexDevices(String type, SubNet sN,
			int x) {
		float energy = 0;
		for (HolonObject hl : sN.getObjects()) {
			float currentEnergyWithoutFlexibles = hl
					.getCurrentEnergyAtTimeStepWithoutFlexiblesAndResetFlexibles(x);

			if (type.equals("prod")) {
				if (currentEnergyWithoutFlexibles > 0) {
					energy += currentEnergyWithoutFlexibles;
					hl.setState(HolonObject.PRODUCER);
				}
			}
			if (type.equals("cons")) {
				if (currentEnergyWithoutFlexibles < 0) {
					energy += currentEnergyWithoutFlexibles;
					hl.setState(HolonObject.NOT_SUPPLIED);
				}
			}
			if (currentEnergyWithoutFlexibles == 0) {
				hl.setState(HolonObject.NO_ENERGY);
			}
		}
		return energy;
	}

	/**
	 * calculates the energy of either all producers or consumers. Flexible
	 * devices are filtered out
	 *
	 * @param type
	 *            Type
	 * @param sN
	 *            Subnet
	 * @param x
	 *            Integer
	 * @return The Energy
	 */
	private float calculateEnergyWithFlexDevices(String type, SubNet sN, int x) {
		float energy = 0;
		for (HolonObject hl : sN.getObjects()) {
			float currentEnergy = hl.getEnergyAtTimeStep(x);

			if (type.equals("prod")) {
				if (currentEnergy > 0) {
					energy += currentEnergy;
					hl.setState(HolonObject.PRODUCER);
				}
			}
			if (type.equals("cons")) {
				if (currentEnergy < 0) {
					energy = energy + currentEnergy;
					hl.setState(HolonObject.NOT_SUPPLIED);
				}
			}
			if (currentEnergy == 0) {
				hl.setState(HolonObject.NO_ENERGY);
			}
		}
		return energy;
	}

	/**
	 * Calculate the Minimum Energy of a Subnet.
	 *
	 * @param sN
	 *            Subnet
	 * @param x
	 *            Integer
	 * @return the Calculated minimum Energy of a Subnet
	 */
	private float calculateMinimumEnergy(SubNet sN, int x) {
		float minimummConsumptionSubnet = 0;
		for (HolonObject hl : sN.getObjects()) {
			float minElement = 0;
			// Search for a activ element
			for (HolonElement he : hl.getElements()) {
				if (he.isActive()) {
					float overallEnergy = he.getOverallEnergyAtTimeStep(x);
					if (overallEnergy < 0) {
						// Is a consumer
						minElement = overallEnergy;
					}
				}
			}
			for (HolonElement he : hl.getElements()) {
				if (he.isActive()) {
					float overallEnergy = he.getOverallEnergyAtTimeStep(x);

					if (minElement < overallEnergy && overallEnergy < 0) {
						// is a smaller consumer
						minElement = overallEnergy;
					}
				}
			}
			minimummConsumptionSubnet += minElement;
		}
		System.out.println("MinimumEnergy = "+ minimummConsumptionSubnet);
		return minimummConsumptionSubnet;
	}

	/**
	 * generates all subNets from all objectsToHandle.
	 */
	private void searchForSubNets() {
		subNets = new ArrayList<>();
		brokenEdges.clear();
		boolean end = false;
		int i = 0;
		AbstractCpsObject cps;
		if (objectsToHandle.size() > 0) {
			while (!end) {
				cps = objectsToHandle.get(i);
				SubNet singleSubNet = new SubNet(new ArrayList<>(),
						new ArrayList<>(), new ArrayList<>(), new ArrayList<>());
				singleSubNet = buildSubNet(cps, new ArrayList<>(), singleSubNet);
				if (singleSubNet.getObjects().size() + singleSubNet.getBatteries().size() != 0 ) {
					subNets.add(singleSubNet);
				}
				if (0 == objectsToHandle.size()) {
					end = true;
				}
			}
		}
	}

	/**
	 * recursivly generates a subnet of all objects, that one specific object is
	 * connected to.
	 *
	 * @param cps
	 *            AbstractCpsObject
	 * @param visited
	 *            visited Array of Integer
	 * @param sN
	 *            Subnets
	 * @return Subnet
	 */
	private SubNet buildSubNet(AbstractCpsObject cps,
			ArrayList<Integer> visited, SubNet sN) {
		visited.add(cps.getId());
		if (cps instanceof HolonObject) {
			sN.getObjects().add((HolonObject) cps);
		}
		if (cps instanceof HolonSwitch) {
			sN.getSwitches().add((HolonSwitch) cps);
		}
		if (cps instanceof HolonBattery) {
			sN.getBatteries().add((HolonBattery) cps);
		}
		removeFromToHandle(cps.getId());
		AbstractCpsObject a;
		AbstractCpsObject b;
		for (CpsEdge edge : cps.getConnections()) {
			if (edge.isWorking()) {
				a = edge.getA();
				b = edge.getB();
				if (!(cps instanceof HolonSwitch)) {
					if (!(sN.getEdges().contains(edge))) {
						sN.getEdges().add(edge);
					}
				}
				if (cps instanceof HolonSwitch
						&& ((HolonSwitch) cps).getState(timeStep)) {
					if (!(sN.getEdges().contains(edge))) {
						sN.getEdges().add(edge);
					}
				}
				if (!visited.contains(a.getId()) && legitState(cps)
						&& !(a instanceof CpsUpperNode)) {
					sN = buildSubNet(a, visited, sN);
				}
				if (!visited.contains(b.getId()) && legitState(cps)
						&& !(b instanceof CpsUpperNode)) {
					sN = buildSubNet(b, visited, sN);
				}
				if (a instanceof CpsUpperNode && a.getId() != cps.getId()) {
					edge.setConnected(CpsEdge.CON_UPPER_NODE_NOT_INSIDE);
					checkForConnectedStates(b, (CpsUpperNode) a, edge);
				}
				if (b instanceof CpsUpperNode && b.getId() != cps.getId()) {
					edge.setConnected(CpsEdge.CON_UPPER_NODE_NOT_INSIDE);
					checkForConnectedStates(a, (CpsUpperNode) b, edge);
				}
			} else {
				brokenEdges.add(edge);
			}
		}
		return sN;
	}

	/**
	 * is the Switch in a legitimate State.
	 *
	 * @param current
	 *            AbstractCpsObject
	 * @return boolean
	 */
	private boolean legitState(AbstractCpsObject current) {
		return !(current instanceof HolonSwitch)
				|| ((HolonSwitch) current).getState(timeStep);
	}

	// /**
	// * ensures that objectsToHandle only contains HolonObjects.
	// */
	// public void cleanObjectsToHandle() {
	// for (int i = 0; i < objectsToHandle.size(); i++) {
	// if (!(objectsToHandle.get(i) instanceof HolonObject)) {
	// objectsToHandle.remove(i);
	// }
	// }
	// }

	/**
	 * removes an Object that already has been handled.
	 *
	 * @param id
	 *            the Object ID
	 */
	private void removeFromToHandle(int id) {
		for (int i = 0; i < objectsToHandle.size(); i++) {
			if (objectsToHandle.get(i).getId() == id) {
				objectsToHandle.remove(i);
			}
		}
	}

	/**
	 * copies the data of an array of Objects.
	 *
	 * @param toCopy
	 *            the ArrayList of CpsObjects co Copy
	 */
	private void copyObjects(ArrayList<AbstractCpsObject> toCopy) {
		for (AbstractCpsObject cps : toCopy) {
			if (cps instanceof CpsUpperNode) {
				copyObjects(((CpsUpperNode) cps).getNodes());
			} else {
				objectsToHandle.add(cps);
			}
		}
	}

	/**
	 * Prints the Components auf all subnets.
	 */
	private void printNetsToConsole() {
		for (int i = 0; i < subNets.size(); i++) {
			SubNet subNet = subNets.get(i);
			System.out.println("SUBNET NR:" + i);
			subNet.toString(timeStep);
		}
	}

	/**
	 * Set the Canvas.
	 *
	 * @param can
	 *            the Canvas
	 */
	public void setCanvas(MyCanvas can) {
		canvas = can;
	}
	/**
	 * Should be a better way to update the canvas -.-
	 * @param can
	 */
	public void updateCanvas() {
		canvas.repaint();
	}
	/**
	 * Reset all Data to the current state of the Model.
	 */
	public void reset() {
		objectsToHandle = new ArrayList<>();
		copyObjects(model.getObjectsOnCanvas());
		flexDevicesTurnedOnThisTurn = new HashMap<>();
	}

	/**
	 * Resets the State of all Edges
	 */
	private void resetEdges() {
		for (CpsEdge e : brokenEdges) {
			e.setWorkingState(true);
		}
	}

	/**
	 * Resets the State for the whole Simulation Model
	 */
	void resetSimulation() {
		reset();
		resetEdges();
	}

	/**
	 * Get all Subnets.
	 *
	 * @return all Subnets
	 */
	public ArrayList<SubNet> getSubNets() {
		return subNets;
	}

	/**
	 * Get broken Edges
	 */
	// public ArrayList<CpsEdge> getBrokenEdges() {
	// return brokenEdges;
	// }

	/**
	 * checks whether a given object is connected to an object inside the
	 * upperNode. if yes, the state for the edge is changed in "connected" or
	 * "not connected"
	 */
	private void checkForConnectedStates(AbstractCpsObject cps,
			CpsUpperNode cUNode, CpsEdge theEdge) {
		AbstractCpsObject tmp;
		for (CpsEdge edge : cps.getConnections()) {
			if (edge.getA().getId() == cps.getId()) {
				tmp = edge.getB();
			} else {
				tmp = edge.getA();
			}
			if (cUNode.getNodes().contains(tmp)) {
				if (tmp instanceof CpsUpperNode) {
					checkForConnectedStates(cps, (CpsUpperNode) tmp, theEdge);
				} else {
					theEdge.setConnected(CpsEdge.CON_UPPER_NODE_AND_INSIDE);
					break;
				}
			}
		}
	}

	public FlexiblePane getFlexiblePane() {
		return flexPane;
	}

	void setFlexiblePane(FlexiblePane fp) {
		flexPane = fp;
	}

	
	public DecoratedState getActualDecorStateWithOffSet(int offSet) {
		return getDecorState(timeStep + offSet);
	}
	public DecoratedState getActualDecorState() {
		return getDecorState(timeStep);
	}
	public VisualRepresentationalState getActualVisualRepresentationalState(){
		return savesVisual.getOrDefault(timeStep, null);
	}
	public DecoratedState getDecorState(int timestep) {
		return saves.getOrDefault(timestep, null);
	}

	public void setGui(GUI gui) {
		this.gui = gui;
	}

	public HashMap<Integer, VisualRepresentationalState> getSavesVisual() {
		return savesVisual;
	}

}
